/*
 *  This is a cleaner plugin module
 *	(c) by kab
 *
 *  It can be compiled by:
 *
 *      - Microsoft Visual C++ .NET 2003
 *
 */
#include "stdafx.h"
#include "cleaner.h"
#include "ver.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

// Global vars:
TiXmlDocument XmlDoc;

typedef CList<CPattern *, CPattern *> PatternsList;

// patterns map
CMapStringToPtr mapPatterns;
// default patterns list
PatternsList lstDefaultPatterns;

// groups map
CMapStringToPtr mapGroups;
// default groups list
CStringList lstDefaultGroups;

// our prefix list
CMapPtrToPtr mapPrefix;
// our postfix list
CMapPtrToPtr mapPostfix;

unsigned long ulFlags=1;

BYTE ToByte(LPCSTR lpStr)
{
	DWORD d;

	if (sscanf(lpStr, "%02X", &d)==0)
		return 0;

	return (BYTE)d;
}

bool CheckSyntax(LPCSTR lpBuf)
{
	if (isxdigit(*lpBuf) && isxdigit(*(lpBuf+1)))
		return true;

	return false;
}

void WriteLog(char *pszName)
{
	DWORD dwTemp;

	HANDLE hFile=CreateFile("cleaner.log", GENERIC_WRITE, FILE_SHARE_WRITE, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL); //   
	if (hFile==INVALID_HANDLE_VALUE) return;
	SetFilePointer(hFile, 0, 0, FILE_END);
	WriteFile(hFile, pszName, strlen(pszName), &dwTemp, NULL);
	CloseHandle(hFile);
}


void ReAnalyse(ea_t addr)
{
	ea_t cur_addr=addr;
	int size;

	if (IsReanalyse(ulFlags))
	{
		while(true)
		{
			size=1;
			do_unknown_range(cur_addr, size, true);
			autoWait();

			while (!ua_code(cur_addr))
			{
				do_unknown_range(cur_addr, size++, true);
				if (size>0x10)
					break;
			}

			autoWait();

			ua_ana0(cur_addr);

			if ((get_byte(cur_addr) & 0xf0)==0x70 ||	// any short Jcc
				(get_byte(cur_addr)==0xff && ((get_byte(cur_addr+1) & 0xf0)==0x80)))	// any near Jcc
				break;	// stop

			cur_addr=get_first_fcref_from(cur_addr) != BADADDR ? get_first_fcref_from(cur_addr) : get_first_cref_from(cur_addr);

			if (cur_addr==BADADDR)
				break;
		}
	}
}

void ReAnalyseSmall(ea_t addr)
{
	if (!isCode(getFlags(addr)))
	{
		do_unknown(addr, true);
		autoWait();
		ua_code(addr);
		autoWait();
	}
}

long ExecuteExpr(ea_t addr, LPCSTR lpExpr, CMapStringToPtr *pMap)
{
	long lRes=0;
	POSITION pos=pMap->GetStartPosition();
	CString strVarName, strVarValue;
	DWORD dwVarValue;
	CString strExrp=lpExpr;
	char errbuf[4096];

	while (pos!=NULL)
	{
		pMap->GetNextAssoc(pos, strVarName, (void*&)dwVarValue);
		strVarValue.Format("0x%X", dwVarValue);
		strExrp.Replace(strVarName, strVarValue);
	}

	strExrp.Replace("SHR", ">>");
	strExrp.Replace("SHL", "<<");
	strExrp.Replace("AND", "&");
	strExrp.Replace("OR",  "|");

	if (IsDebugMode(ulFlags))
	{
		if (IsWriteLog(ulFlags))
		{
			WriteLog("ExecuteExpr: ");
			WriteLog((char *)(LPCTSTR)strExrp);
			WriteLog("\n");
		}
		else
			msg("%s\n",strExrp);
	}


	if (!calcexpr_long(addr, strExrp, &lRes, errbuf, 4096))
	{
		msg(errbuf);
		return 0;
	}

	return lRes;
}

int CalcCounter(LPCSTR lpMask, int &iRes, CMapStringToPtr *pMap)
{
	char buf[MAX_PATH];
	int iCnt=0;

	memset(buf, 0, MAX_PATH);
	iRes=0;

	if (lpMask[iCnt]=='(')
	{
		iCnt++; // (
		iCnt+=ParseExpr(lpMask+iCnt, buf, ')');
		iCnt++; // )

		iRes=ExecuteExpr(BADADDR, buf, pMap);
	}

	return iCnt;
}

long ParseExpr(LPCSTR lpInBuf, LPSTR lpOutBuf, char cStopChar)
{
	long i=0;

	for (i=0; lpInBuf[i]!=cStopChar && lpInBuf[i]!=0; i++)
		lpOutBuf[i]=lpInBuf[i];

	return i;
}

bool CalcExprAndCompare(ea_t addr, LPCSTR lpMask, UINT &iMaskLen, UINT &iInstrLen, CMapStringToPtr *pMap)
{
	char buf[MAX_PATH];
	memset(buf, 0, MAX_PATH);

	LPCSTR lpExpr=NULL;
	LONG lRes=0;

	iInstrLen=0;
	iMaskLen=0;

	if (lpMask[iMaskLen]=='(')
	{
		iMaskLen++;		// '('
		iMaskLen+=ParseExpr(lpMask+iMaskLen, buf, ')');
		iMaskLen++;		// ')'

		lpExpr=strtok(buf, ":");

		LPCSTR lpLen=strtok(NULL, ":");
		if (lpLen)
			iInstrLen=atoi(lpLen);
		else
			iInstrLen=4;

		lRes=ExecuteExpr(addr, lpExpr, pMap);

		if (IsDebugMode(ulFlags))
		{
			if (IsWriteLog(ulFlags))
			{
				char tmpbuf1[4096];
				qsnprintf(tmpbuf1, 4096, "CalcExprAndCompare Result=%X\n", lRes);
				WriteLog(tmpbuf1);
			}
			else
				msg("CalcExprAndCompare Result=%X\n", lRes);
		}

		switch(iInstrLen)
		{
		case 1:
			if (get_byte(addr)==LOBYTE(LOWORD(lRes)))
				return true;
			break;
		case 2:
			if (get_word(addr)==LOWORD(lRes))
				return true;
			break;
		case 4:
			if (get_long(addr)==lRes)
				return true;
			break;
		default:
			msg("Unsupported length of expression: %X\n", iInstrLen);
			break;
		}
	}

	return false;
}

int CalcExprAndWrite(ea_t addr, LPCSTR lpMask, UINT &iLen, CMapStringToPtr *pMap)
{
	char buf[MAX_PATH];
	memset(buf, 0, MAX_PATH);
	LPCSTR lpExpr=NULL;
	LONG lRes=0;

	UINT iStrLen=0;

	if (lpMask[iStrLen]=='[')
	{
		iStrLen++;		// '['
		iStrLen+=ParseExpr(lpMask+iStrLen, buf, ']');
		iStrLen++;		// ']'

		lpExpr=strtok(buf, ":");

		LPCSTR lpLen=strtok(NULL, ":");
		if (lpLen)
			iLen=atoi(lpLen);
		else
			iLen=4;

		lRes=ExecuteExpr(addr, lpExpr, pMap);

		if (IsDebugMode(ulFlags))
		{
			if (IsWriteLog(ulFlags))
			{
				char tmpbuf1[4096];
				qsnprintf(tmpbuf1, 4096, "CalcExprAndWrite Result=%X\n", lRes);
				WriteLog(tmpbuf1);
			}
			else
				msg("CalcExprAndWrite Result=%X\n", lRes);
		}
        
		switch(iLen)
		{
		case 1:
			put_byte(addr, lRes);
			break;
		case 2:
			put_word(addr, lRes);
			break;
		case 4:
			put_long(addr, lRes);
			break;
		default:
			msg("Unsupported length of expression: %X\n", iLen);
			break;
		}
	}

	return iStrLen;
}

long CalcExprAndCheck(ea_t addr, LPCSTR lpMask, UINT &iLen, CMapStringToPtr *pMap)
{
	char buf[MAX_PATH];
	LONG lRes=0;

	iLen=0;

	memset(buf, 0, MAX_PATH);

	if (lpMask[iLen]=='[')
	{
		iLen++;		// '['
		iLen+=ParseExpr(lpMask+iLen, buf, ']');
		iLen++;		// ']'

		lRes=ExecuteExpr(addr, buf, pMap);

		if (IsDebugMode(ulFlags))
		{
			if (IsWriteLog(ulFlags))
			{
				char tmpbuf1[4096];
				qsnprintf(tmpbuf1, 4096, "CalcExprAndCheck Result=%X\n", lRes);
				WriteLog(tmpbuf1);
			}
			else
				msg("CalcExprAndCheck Result=%X\n", lRes);
		}
	}

	return lRes;
}


int ShowExpr(ea_t addr, LPCSTR lpMask, CMapStringToPtr *pMap)			// just calc and out expression
{
	char buf[MAX_PATH];
	memset(buf, 0, MAX_PATH);
	LONG lRes=0;

	UINT iStrLen=0;

	if (lpMask[iStrLen]=='[')
	{
		iStrLen++;		// '['
		iStrLen+=ParseExpr(lpMask+iStrLen, buf, ']');
		iStrLen++;		// ']'

		lRes=ExecuteExpr(addr, buf, pMap);
		msg("Result=%X at Address=%X\n", lRes, addr);
	}

	return iStrLen;
}

ea_t SkipInstruction(ea_t addr)
{
	ea_t cur_addr=get_item_head(addr);

	if (cur_addr!=BADADDR)
	{
		ua_ana0(cur_addr);
		return cur_addr+cmd.size;
	}

	return addr;
}

int JumpExpr(ea_t &addr, LPCSTR lpMask, CMapStringToPtr *pMap)			// jump to expression
{
	char buf[MAX_PATH];
	memset(buf, 0, MAX_PATH);
	LONG lRes=0;

	UINT iStrLen=0;

	if (lpMask[iStrLen]=='[')
	{
		iStrLen++;		// '['
		iStrLen+=ParseExpr(lpMask+iStrLen, buf, ']');
		iStrLen++;		// ']'

		lRes=ExecuteExpr(addr, buf, pMap);
		addr=lRes;
		if (IsDebugMode(ulFlags))
		{
			if (IsWriteLog(ulFlags))
			{
				char tmpbuf1[4096];
				qsnprintf(tmpbuf1, 4096, "Jump to: %X\n", addr);
				WriteLog(tmpbuf1);
			}
			else
				msg("Jump to: %X\n", addr);
		}
	}

	return iStrLen;
}

ea_t SkipSingleSearch(ea_t addr, CString strSearch)
{
	if (IsMask(addr, strSearch, NULL)!=BADADDR)
	{
		ua_code(addr);
		return (get_first_fcref_from(addr) == BADADDR) ? 
			get_first_cref_from(addr) : get_first_fcref_from(addr);
	}
	return BADADDR;
}

ea_t SkipSingleGroup(ea_t addr, CStringList *plst)
{
	CString str;
	POSITION pos;
	ea_t res=BADADDR;

	pos=plst->GetHeadPosition();
	for (int i=0; i<plst->GetSize(); i++)
	{
		str=plst->GetNext(pos);

		res=SkipSingleSearch(addr, str);

		if (res!=BADADDR)
			break;
	}

	return res;
}

bool SkipGroup(ea_t &addr, bool bIsMultiple, LPCSTR lpBuf, UINT &iLen)
{
	ea_t cur_addr=addr, last_addr=addr;
	CStringList lst, *pLstGroup;
	CString str;
	int iCnt=0;

	iLen=SplitString(lpBuf, &lst);

	while (true)
	{
		if (!lst.IsEmpty())			// named groups
		{
			POSITION pos;
			pos=lst.GetHeadPosition();
			for (int i=0; i<lst.GetSize(); i++)
			{
				str=lst.GetNext(pos);
				if (!mapGroups.Lookup(str, (void *&)pLstGroup))
				{
					warning("Group %s not found\n", str);
					return false;
				}

				cur_addr=SkipSingleGroup(cur_addr, pLstGroup);

				if (cur_addr!=BADADDR)
					break;
			}
		}
		else						// default groups
			cur_addr=SkipSingleGroup(cur_addr, &lstDefaultGroups);

		if (cur_addr==BADADDR)
			break;

		last_addr=cur_addr;

		if (!bIsMultiple)
			break;

		if (iCnt++>256)
		{
			msg("Error: too deep recursion for group skipping\n");
			return false;
		}
	}

	if (last_addr==addr)		// no jumps performed
		return false;

	addr=last_addr;
	return true;
}

bool SkipPattern(ea_t &addr, bool bIsMultiple, LPCSTR lpBuf, UINT &iLen)
{
	ea_t cur_addr=addr, last_addr=addr;
	CStringList lst;
	CString str;
	CPattern *p;
	POSITION pos;

	iLen=SplitString(lpBuf, &lst);

	while (true)
	{
		if (!lst.IsEmpty())		// named patterns
		{
			pos=lst.GetHeadPosition();
			for (int i=0; i<lst.GetSize(); i++)
			{
				str=lst.GetNext(pos);
				if (!mapPatterns.Lookup(str, (void *&)p))
				{
					warning("Pattern %s not found\n", str);
					return false;
				}

				cur_addr=Action(cur_addr, p);

				if (cur_addr!=BADADDR)
					break;
			}
		}
		else				// default patterns
		{
			pos=lstDefaultPatterns.GetHeadPosition();
			for (int i=0; i<lstDefaultPatterns.GetSize(); i++)
			{
				p=lstDefaultPatterns.GetNext(pos);

				cur_addr=Action(cur_addr, p);

				if (cur_addr!=BADADDR)
					break;
			}
		}

		if (cur_addr==BADADDR)
			break;

		last_addr=cur_addr;

		if (!bIsMultiple)
			break;
	}

	if (last_addr==addr)		// no jumps performed
		return false;

	addr=last_addr;
	return true;
}

UINT SplitString(LPCSTR lpBuf, CStringList *lst)
{
	CString str;
	UINT iLen=0;

	str.Empty();

	if (lpBuf[iLen]=='[')
	{
		for (iLen=1; lpBuf[iLen]!=']'; iLen++)
			str+=lpBuf[iLen];

		iLen++;

		int curPos=0;
		CString resToken;

		resToken=str.Tokenize(",",curPos);
		while (resToken != "")
		{
			lst->AddTail(resToken);
			resToken=str.Tokenize(",",curPos);
		}
	}

	return iLen;
}

int FillVars(ea_t addr, LPCSTR lpMask, UINT &iLen, CMapStringToPtr *pMap)
{
	char buf[MAX_PATH];
	memset(buf, 0, MAX_PATH);

	UINT iStrLen=0;
	LPCSTR lpVar=NULL;
	ULONG l=0;

	if (lpMask[iStrLen]=='[')
	{
		iStrLen++;		// '['
		iStrLen+=ParseExpr(lpMask+iStrLen, buf, ']');
		iStrLen++;		// ']'

		lpVar=strtok(buf, ":");

		LPCSTR lpLen=strtok(NULL, ":");
		if (lpLen)
			iLen=atoi(lpLen);
		else
			iLen=4;

		switch(iLen)
		{
		case 1:
			l=get_byte(addr);
			break;
		case 2:
			l=get_word(addr);
			break;
		case 4:
			l=get_long(addr);
			break;
		default:
			msg("Unsupported length of var: %X\n", iLen);
			break;
		}

		pMap->SetAt(lpVar, (void*)l);
	}

	return iStrLen;
}

int SetVars(ea_t addr, LPCSTR lpMask, CMapStringToPtr *pMap)
{
	char buf[MAX_PATH];
	memset(buf, 0, MAX_PATH);

	UINT iStrLen=0;

	if (lpMask[iStrLen]=='[')
	{
		iStrLen++;		// '['
		iStrLen+=ParseExpr(lpMask+iStrLen, buf, ']');
		iStrLen++;		// ']'

		pMap->SetAt(buf, (void*)addr);
	}

	return iStrLen;
}

ea_t get_first_cref_to_high( ea_t ea ) 
{ 
	ea_t xref = get_first_cref_to( ea ); 

	if( xref == BADADDR ) 
	{ 
		return xref; 
	} 
	else if( xref + get_item_size( xref ) == ea ) 
	{ 
		// continue search 
		return get_next_cref_to( ea, xref ); 
	} 
	else 
	{ 
		return xref; 
	} 
} 

void HideByte(ea_t addr, bool bStrong)
{
	bool bReferenced=false;

	// unhide first
	unhide_item( addr ); 

	// get_first_dref_to works only with data 
	bReferenced = ( get_first_dref_to( addr ) != BADADDR ); 

	// make code 
	ua_code( addr ); 

	// check if any xref 
	bReferenced |= ( get_first_cref_to_high( addr ) != BADADDR ); 

	if( !bStrong || !bReferenced ) 
	{ 
		// hide current NOP 
		hide_item( addr ); 
	} 
}
int CalcHide(LPCSTR lpMask, int &iRes, unsigned char &bByte, CMapStringToPtr *pMap)
{
	char buf[MAX_PATH];
	int iCnt=0;

	memset(buf, 0, MAX_PATH);
	iRes=0;
	bByte=0;

	if (lpMask[iCnt]=='(')
	{
		iCnt++;
		for (int i=0; lpMask[iCnt]!=')' && lpMask[iCnt]!=',' && lpMask[iCnt]!='\x0'; i++, iCnt++)
			buf[i]=lpMask[iCnt];


		iRes=ExecuteExpr(BADADDR, buf, pMap);

		if (lpMask[iCnt]==',')
		{
			iCnt++;
			memset(buf, 0, MAX_PATH);
			for (int i=0; lpMask[iCnt]!=')' && lpMask[iCnt]!=',' && lpMask[iCnt]!='\x0'; i++, iCnt++)
				buf[i]=lpMask[iCnt];
			
			bByte=(unsigned char)ExecuteExpr(BADADDR, buf, pMap);
		}

		iCnt++;
	}

	return iCnt;
}

int HideItem(ea_t &addr, LPCSTR lpMask, CMapStringToPtr *pMap, bool bStrong)
{
	int iRes=0;
	int iCnt=0;
	unsigned char b=0;

	if (*lpMask=='(')	
	{
		iRes=CalcHide(lpMask, iCnt, b, pMap);
		for (int i=0; i<iCnt; i++)
		{
			if (b)
			{
				do_unknown(addr, false);
				put_byte(addr, b);
			}
			HideByte(addr, bStrong);
			addr++;
		}
	}
	else	// just 1 byte
	{
		HideByte(addr, bStrong);
		addr++;
	}

	return iRes;
}

ea_t Replace(ea_t addr, LPCSTR lpReplace, CMapStringToPtr *pMap)
{
	UINT i=0, iLen=0;
	ea_t cur_addr=addr;

	while (i<strlen(lpReplace))
	{
		if (pMap)
			pMap->SetAt("Current", (void*)cur_addr);

		if ( (lpReplace[i]=='-') && (lpReplace[i+1]=='-') )
		{
			i+=2;
			cur_addr++;
		}
		else if (lpReplace[i]=='-' && lpReplace[i+1]=='c')
		{
			i+=2;
			MakeComment(cur_addr, false, lpReplace+i, iLen);
			i+=iLen;
		}
		else if (lpReplace[i]=='-' && lpReplace[i+1]=='C')
		{
			i+=2;
			MakeComment(cur_addr, true, lpReplace+i, iLen);
			i+=iLen;
		}
		else if (lpReplace[i]=='-' && lpReplace[i+1]=='j' && pMap)
		{
			i+=2;
			SkipGroup(cur_addr, false, lpReplace+i, iLen);
			i+=iLen;
		}
		else if (lpReplace[i]=='-' && lpReplace[i+1]=='J' && pMap)
		{
			i+=2;

			if (!SkipGroup(cur_addr, false, lpReplace+i, iLen))
				return false;

			i+=iLen;
		}
		else if (lpReplace[i]=='-' && lpReplace[i+1]=='f' && pMap)
		{
			i+=2;
			SkipGroup(cur_addr, true, lpReplace+i, iLen);
			i+=iLen;
		}
		else if (lpReplace[i]=='-' && lpReplace[i+1]=='F' && pMap)
		{
			i+=2;

			if (!SkipGroup(cur_addr, true, lpReplace+i, iLen))
				return false;

			i+=iLen;
		}
		else if (lpReplace[i]=='-' && lpReplace[i+1]=='i')	// skip instr
		{
			ReAnalyse(addr);
			i+=2;
			cur_addr=SkipInstruction(cur_addr);
		}
		else if (lpReplace[i]=='-' && lpReplace[i+1]=='s' && pMap)	// just search
		{
			i+=2;
			i+=ShowExpr(cur_addr, lpReplace+i, pMap);
		}
		else if (lpReplace[i]=='-' && lpReplace[i+1]=='g' && pMap)	// jump to expression
		{
			i+=2;
			i+=JumpExpr(cur_addr, lpReplace+i, pMap);
		}
		else if (lpReplace[i]=='-' && lpReplace[i+1]=='m' && pMap)	// mark checkpoint
		{
			i+=2;
			i+=SetVars(cur_addr, lpReplace+i, pMap);
		}
		else if (lpReplace[i]=='[' && pMap)
		{
			i+=CalcExprAndWrite(cur_addr, lpReplace+i, iLen, pMap);
			cur_addr+=iLen;
		}
		else if (lpReplace[i]=='-' && lpReplace[i+1]=='h' && pMap)	// hide byte
		{
			i+=2;
			i+=HideItem(cur_addr, lpReplace+i, pMap, false);
		}
		else if (lpReplace[i]=='-' && lpReplace[i+1]=='H' && pMap)	// hide byte strong
		{
			i+=2;
			i+=HideItem(cur_addr, lpReplace+i, pMap, true);
		}
		else		// Usual byte
		{
			if (!CheckSyntax(lpReplace+i))
			{
				msg("Syntax error!\n");
				msg("Replace pattern: %s, Position: %d, Address: 0x%X\n", lpReplace, i, cur_addr);
				return false;
			}

			BYTE b=ToByte(lpReplace+i);
			i+=2;

			if (lpReplace[i]=='(')
			{
				int iCnt=0;
				i+=CalcCounter(lpReplace+i, iCnt, pMap);

				if (iCnt>0)
					for (; iCnt; iCnt--)
					{
						put_byte(cur_addr, b);
						cur_addr++;
					}
			}
			else
			{
				put_byte(cur_addr, b);
				cur_addr++;
			}
		}
	}

	ReAnalyse(addr);

	return cur_addr;
}

bool MakeComment(ea_t &addr, bool bIsRepeatable, LPCSTR lpBuf, UINT &iLen)
{
	char buf[MAX_PATH];
	
	iLen=0;

	if (lpBuf[iLen]=='[')
	{
		iLen++;		// '['
		for (int i=0; lpBuf[iLen]!=']'; i++, iLen++)
			buf[i]=lpBuf[iLen];
		iLen++;		// ']'

		return set_cmt(addr, buf, bIsRepeatable);
	}

	return false;
}

ea_t IsMask(ea_t addr, LPCSTR lpMask, CMapStringToPtr *pMap)
{
	UINT i=0, iLen=0, iInstrLen=0;
	ea_t cur_addr=addr;

	if (pMap)
	{
		pMap->RemoveAll();
		pMap->SetAt("Begin", (void*)cur_addr);
	}
	
	while (i<strlen(lpMask))
	{
		iLen=0;

		if (pMap)
			pMap->SetAt("Current", (void*)cur_addr);

		if ( (lpMask[i]=='-') && (lpMask[i+1]=='-') )
		{
			i+=2;
			cur_addr++;
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='c')
		{
			i+=2;
			MakeComment(cur_addr, false, lpMask+i, iLen);
			i+=iLen;
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='C')
		{
			i+=2;
			MakeComment(cur_addr, true, lpMask+i, iLen);
			i+=iLen;
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='j' && pMap)
		{
			i+=2;
			SkipGroup(cur_addr, false, lpMask+i, iLen);
			i+=iLen;
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='J' && pMap)
		{
			i+=2;

			if (!SkipGroup(cur_addr, false, lpMask+i, iLen))
				return BADADDR;

			i+=iLen;
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='f' && pMap)
		{
			i+=2;
			SkipGroup(cur_addr, true, lpMask+i, iLen);
			i+=iLen;
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='F' && pMap)
		{
			i+=2;

			if (!SkipGroup(cur_addr, true, lpMask+i, iLen))
				return BADADDR;

			i+=iLen;
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='p' && pMap)	// skip pattern
		{
			i+=2;
			SkipPattern(cur_addr, false, lpMask+i, iLen);
			i+=iLen;
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='P' && pMap)	// skip pattern strong
		{
			i+=2;

			if (!SkipPattern(cur_addr, false, lpMask+i, iLen))
				return BADADDR;

			i+=iLen;
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='a' && pMap)	// skip a lot of patterns
		{
			i+=2;
			SkipPattern(cur_addr, true, lpMask+i, iLen);
			i+=iLen;
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='A' && pMap)	// skip a lot of patterns strong
		{
			i+=2;

			if (!SkipPattern(cur_addr, true, lpMask+i, iLen))
				return BADADDR;

			i+=iLen;
		}
		else if (lpMask[i]=='[' && pMap)						// store variable
		{
			i+=FillVars(cur_addr, lpMask+i, iInstrLen, pMap);
			cur_addr+=iInstrLen;
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='s' && pMap)	// just search
		{
			i+=2;
			i+=ShowExpr(cur_addr, lpMask+i, pMap);
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='g' && pMap)	// jump to expression
		{
			i+=2;
			i+=JumpExpr(cur_addr, lpMask+i, pMap);
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='m' && pMap)	// mark checkpoint
		{
			i+=2;
			i+=SetVars(cur_addr, lpMask+i, pMap);
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='i')			// skip instr
		{
			i+=2;
			cur_addr=SkipInstruction(cur_addr);
		}
		else if (lpMask[i]=='(' && pMap)						// check expression match
		{
			if (!CalcExprAndCompare(cur_addr, lpMask+i, iLen, iInstrLen, pMap))
				return BADADDR;

			i+=iLen;
			cur_addr+=iInstrLen;
		}
		else if (lpMask[i]=='-' && lpMask[i+1]=='=' && pMap)	// execute bool expression
		{
			i+=2;
			if (!CalcExprAndCheck(cur_addr, lpMask+i, iLen, pMap))
				return BADADDR;
			i+=iLen;
		}
		else		// Usual byte
		{
			BYTE b1=get_byte(cur_addr);

			if (!CheckSyntax(lpMask+i))
			{
				msg("Syntax error!\n");
				msg("Search pattern: %s, Position: %d, Address: 0x%X\n", lpMask, i, cur_addr);
				return BADADDR;
			}

			BYTE b2=ToByte(lpMask+i);
			BYTE b3=0xff;

			i+=2;

			if (lpMask[i]==':')		// bit mask
			{
				i++;

				if (!CheckSyntax(lpMask+i))
				{
					msg("Syntax error!\n");
					msg("Search pattern: %s, Position: %d, Address: 0x%X\n", lpMask, i, cur_addr);
					return BADADDR;
				}

				b3=ToByte(lpMask+i);
				i+=2;
			}

			if ( (b1 & b3) != (b2 & b3) )
				return BADADDR;

			cur_addr++;
		}
	}

	if (pMap)
		pMap->SetAt("End", (void*)cur_addr);

	return cur_addr;
}

ea_t Action(ea_t start_addr, CPattern *p)
{
	CMapStringToPtr mapVars;
	ea_t addr=BADADDR;
	static int iDepth=0;

	if (wasBreak())
		return BADADDR;

	if (iDepth++>256)
	{
		msg("Error: too deep recursion for pattern: %s\n", p->GetName());
		return BADADDR;
	}

	if (!p->GetSearchMask().IsEmpty())
	{
		addr=IsMask(start_addr, p->GetSearchMask(), &mapVars);
		if (addr!=BADADDR)
		{
			if (!p->GetReplaceMask().IsEmpty())
				Replace(start_addr, p->GetReplaceMask(), &mapVars);
//				addr=Replace(start_addr, p->GetReplaceMask(), &mapVars);
		}
	}

	iDepth--;

	return addr;
}

bool Apply(ea_t start_addr)
{
	bool bRes=false;
	POSITION pos;
	CPattern *p;
	CString str;

	try
	{
		if (IsSkipNonCode(ulFlags))				// check for code
		{
			flags_t f=getFlags(start_addr);

			if (isTail(f))
				f=getFlags(get_item_head(start_addr));

			if (!isCode(f))
				return false;
		}

		if (wasBreak())
		{
			hide_wait_box();
			return false;
		}

		pos=mapPatterns.GetStartPosition();
		while (pos!=NULL)
		{
			if (wasBreak())
				return false;

			mapPatterns.GetNextAssoc(pos, str, (void *&)p);

			if (IsDebugMode(ulFlags))
			{
				char buf[4096];

				qsnprintf(buf, 4096, "Trying pattern %s at %X\n", p->GetName(), start_addr);

				if (IsWriteLog(ulFlags))
					WriteLog(buf);
				else
					msg(buf);
			}

			if (p && Action(start_addr, p)!=BADADDR)
			{
				msg("Pattern %s found at %X\n", p->GetName(), start_addr);
				bRes=true;
			}
		}
	}
	catch(...)
	{
		msg("Internal plugin error at: %X\n", start_addr);
		bRes=false;
	}

	return bRes;
}

ea_t SkipJump(ea_t addr)
{
	ea_t res=addr;

	while (get_byte(res)==0xe9 || get_byte(res)==0xeb || get_byte(res)==0x90)
	{
		ua_code(res);
		if (get_first_fcref_from(res)!=BADADDR)
			res=get_first_fcref_from(res);
		else
			res=get_first_cref_from(res);
	}

	if (res==BADADDR)
        warning("SkipJump failed!");

	return res;
}
BYTE IsConditionalJump(ea_t addr)
{
	if (get_byte(addr)==0xe3)
		return get_byte(addr);
	if ((get_byte(addr) & 0xf0)==0x70)
		return get_byte(addr);
	if (get_byte(addr)==0x0f && ((get_byte(addr+1) & 0xf0)==0x80))
		return (get_byte(addr+1)-0x10);

	return 0;
}

void DFS_WORK_TOPOLOGY(int iVal, LPBYTE lpVal, C2DArray *arr, CStack *stack)
{
	lpVal[iVal]=1;
	for (UINT i=0;i<arr->GetSize();i++)
	{
		if ((arr->GetAt(iVal, i)) && (!lpVal[i]))
			DFS_WORK_TOPOLOGY(i, lpVal, arr, stack);
	}
	stack->Push(iVal);
	//msg("%X\n", iVal);
}

UINT BuildGraph(ea_t cur_addr, ea_t from_addr, C2DArray *arr, CArray<ea_t, const ea_t&> *arrDone)
{
	UINT iCol=0, iRow=0;
	UINT iCnt;
	ea_t to_addr=BADADDR;
	int iRes=-1;
	BYTE bJump=0;
	bool bExist=false;
	CString *s1, *s2;

	while (cur_addr!=BADADDR)
	{
		ReAnalyseSmall(cur_addr);

		if (wasBreak())
		{
			hide_wait_box();
			return -1;
		}

		if (IsWriteLog(ulFlags))
		{
			char tmpbuf1[4096];
			qsnprintf(tmpbuf1, 4096, "BuildGraph: start at %X\n", cur_addr);
			WriteLog(tmpbuf1);
		}

		s1=new CString;
		s1->Format("loc_%X:\n", cur_addr);

		cur_addr=SkipJump(cur_addr);			//   JUMP-

		if (cur_addr==BADADDR)
			return -1;

		ReAnalyseSmall(cur_addr);

		if (mapPrefix.Lookup((void *)cur_addr, (void *&)s2))
		{
			mapPrefix.RemoveKey((void *)cur_addr);
			delete s2;
		}

		mapPrefix[(void *)cur_addr]=s1;

		if (IsWriteLog(ulFlags))
		{
			char tmpbuf1[4096];
			qsnprintf(tmpbuf1, 4096, "BuildGraph: After SkipJump we at %X\n", cur_addr);
			WriteLog(tmpbuf1);
		}

		bJump=IsConditionalJump(cur_addr);		//   Jcc?

		if (bJump)
		{
			ReAnalyseSmall(get_first_fcref_from(cur_addr));
			iCol=BuildGraph(get_first_fcref_from(cur_addr), from_addr, arr, arrDone);

			for (int i=0; i<arrDone->GetSize(); i++)
				if (arrDone->GetAt(i)==from_addr)
				{
					iRow=i;
					break;
				}

			if (iRow!=iCol)
				arr->SetAt(iRow, iCol, bJump);


			ReAnalyseSmall(get_first_cref_from(cur_addr));
			to_addr=get_first_cref_from(cur_addr);
			if (to_addr==BADADDR)
				return -1;

			while (to_addr==get_first_fcref_from(cur_addr))
			{
				if (wasBreak())
				{
					hide_wait_box();
					return -1;
				}

				to_addr=get_next_cref_from(cur_addr, to_addr);

				if (to_addr==BADADDR)
					return -1;
			}

			cur_addr=to_addr;
		}
		else									//  
		{
			for (iCnt=0;(int)iCnt<arrDone->GetSize();iCnt++)		// check that item already exist
			{
				if (arrDone->GetAt(iCnt)==cur_addr)
				{
					bExist=true;
					break;
				}
			}

			if (bExist)										// exist? make ref
			{
				iCol=iCnt;
				for (int i=0;i<arrDone->GetSize();i++)
					if (arrDone->GetAt(i)==from_addr)
					{
						iRow=i;
						break;
					}

				if (iRow!=iCol)
				{
					arr->SetAt(iRow, iCol, 1);
					if (iRes==-1)
						iRes=iCnt;
				}
				
				return iRes;
			}
			else											// add instruction
			{
				iCol=arrDone->Add(cur_addr);

				if (iRes==-1)
					iRes=iCol;
				
				if (IsDebugMode(ulFlags))
				{
					s1=new CString;
					s1->Format("0x%X (%X)", cur_addr, iCol);

					if (mapPostfix.Lookup((void *)cur_addr, (void *&)s2))
					{
						mapPostfix.RemoveKey((void *)cur_addr);
						delete s2;
					}

					mapPostfix[(void *)cur_addr]=s1;
				}

				for (int i=0;i<arrDone->GetSize();i++)
					if (arrDone->GetAt(i)==from_addr)
					{
						iRow=i;
						break;
					}

				if (iRow!=iCol)
					arr->SetAt(iRow, iCol, 1);

				from_addr=cur_addr;
				cur_addr=get_first_cref_from(cur_addr);
			}
		}
	}

	return iRes;
}

const char * GetJcc(BYTE bVal)
{
	return JCC[bVal-0x70];
}

bool RebuildCode(ea_t addr, LPSTR lpFilename)
{
	ea_t cur_addr=addr;

	CArray<ea_t, const ea_t&> arrDone;		// done instructions

	C2DArray arr;				//   
	LPBYTE lpValues;			//    

	CStack stack;				//    

	UINT iVal=0;					// current value
	UINT iTarget=0;				// target value
	UINT iPrev=0;				// previous value

	BYTE bJump=0;

	char buf[1024];				// buffer for output

	CFile fOut;

	CleanUpLists();

	if (!fOut.Open(lpFilename, CFile::modeCreate | CFile::modeNoTruncate | CFile::modeWrite))
		return false;

	fOut.SeekToEnd();

	if (BuildGraph(addr, addr, &arr, &arrDone)==-1)
		return false;

	if (wasBreak())
	{
		hide_wait_box();
		return false;
	}

	if (IsDebugMode(ulFlags))
	{
		fOut.Write("; Builded graph\n", strlen("; Builded graph\n"));
		for (UINT i=0;i<arr.GetSize();i++)
		{
			fOut.Write("; ", strlen("; "));
			for (UINT j=0;j<arr.GetSize();j++)
			{
				qsnprintf(buf, 1024, "%02X ", arr.GetAt(i, j));
				fOut.Write(buf, strlen(buf));
			}
			fOut.Write("\n", strlen("\n"));
		}
	}

	lpValues=new BYTE[arr.GetSize()];

	memset(lpValues, 0, arr.GetSize());

	for (UINT i=0;i<arr.GetSize();i++)
		if (lpValues[i]!=1)
			DFS_WORK_TOPOLOGY(i, lpValues, &arr, &stack);

	delete [] lpValues;

	if (IsDebugMode(ulFlags))
	{
		fOut.Write("; Output order:\n", strlen("; Output order:\n"));
		fOut.Write("; ", strlen("; "));
		for (POSITION pos=stack.GetHeadPosition(); pos!=NULL; stack.GetNext(pos))
		{
			qsnprintf(buf, 1024, "%X ", stack.GetAt(pos));
			fOut.Write(buf, strlen(buf));
		}
		fOut.Write("\n", strlen("\n"));
	}

	while (get_first_cref_from(cur_addr=arrDone.GetAt(stack.GetTail()))!=BADADDR)
	{
		iVal=stack.RemoveTail();
		iTarget=0;
		for (iTarget=0; iTarget<arr.GetSize(); iTarget++)
			if (arr.GetAt(iVal, iTarget)==1)
				break;

		for (POSITION pos = stack.GetHeadPosition(); pos!=NULL; stack.GetNext(pos))
		{
			if (stack.GetAt(pos)==iTarget)
			{
				stack.InsertBefore(pos, iVal);
				break;
			}
		}
	}

	iVal=0;
	iTarget=0;
	iPrev=0;

	while ( stack.Peek()!=BADADDR )
	{
		iVal=stack.Pop();
		iTarget=stack.Peek();

		cur_addr=arrDone.GetAt(iVal);

		if ((arr.GetAt(iPrev, iVal)!=1) ||	//   
			(arr.SumCol(iVal)>1))			//  1    
		{
			//  

			//get_name(BADADDR, cur_addr, buf, 1024);
			//msg(buf);
			//msg(":\n");

			CString *s;
			if (mapPrefix.Lookup((void *)cur_addr, (void *&)s))
				fOut.Write(*s, s->GetLength());

			qsnprintf(buf, 1024, "loc_%X:\n", cur_addr);
			fOut.Write(buf, strlen(buf));
		}

		generate_disasm_line(cur_addr, buf, 1024, 0);
		tag_remove(buf, buf, 1024);
		CString *s;
		if (mapPostfix.Lookup((void *)cur_addr, (void *&)s))
		{
			while (strlen(buf)<8*6)
				qstrncat(buf, " ", 1024);

			qstrncat(buf, "\t;", 1024);
			qstrncat(buf, *s, 1024);
		}
		qstrncat(buf, "\n", 1024);

		//msg(buf);
		fOut.Write("\t", strlen("\t"));
		fOut.Write(buf, strlen(buf));

		for (UINT i=0; i<arr.GetSize(); i++)		//    JUMP-
		{
			if (i!=iVal)							//   
			{
				if (arr.GetAt(iVal, i)>1)			//  Jcc
				{
                    //msg(GetJcc(arr.GetAt(iVal, i)));
					fOut.Write("\t", strlen("\t"));
					fOut.Write(GetJcc(arr.GetAt(iVal, i)), strlen(GetJcc(arr.GetAt(iVal, i))));

					//get_name(BADADDR, arrDone.GetAt(i), buf, 1024);
					//msg(buf);
					//msg("\n");
					qsnprintf(buf, 1024, "loc_%X\n", arrDone.GetAt(i));
					fOut.Write(buf, strlen(buf));
				}
			}
		}

		if (iTarget!=BADADDR)				// not end
		{
			if (arr.GetAt(iVal, iTarget)!=1)
			{
				for (UINT i=0; i<arr.GetSize(); i++)
					if (arr.GetAt(iVal, i)==1)
					{
						//msg("jmp     ");
						//get_name(BADADDR, arrDone.GetAt(i), buf, 1024);
						//msg(buf);
						//msg("\n");
						fOut.Write("\tjmp     ", strlen("\tjmp     "));
						qsnprintf(buf, 1024, "loc_%X\n", arrDone.GetAt(i));
						fOut.Write(buf, strlen(buf));
						break;
					}
			}
		}

		iPrev=iVal;
	}

	arr.RemoveAll();

	fOut.Close();

	return true;
}

CString ParseElement(TiXmlNode *xmlnode)
{
	TiXmlNode *xmlchild;
	CString str="";

	if (xmlnode)
	{
		xmlchild = NULL;

		while (xmlchild = xmlnode->IterateChildren (xmlchild)) 
			if (xmlchild->Type() == TiXmlNode::TEXT ) str += xmlchild->Value();

		str.Remove(' ');
		str.Remove('\x0D');
		str.Remove('\x0A');
		str.Remove('\t');
	}

	return str;
}

bool Parse(void)
{
	TiXmlNode *xmlpattern, *xmlgroup, *xmlapp, *xmlsearch, *xmlrepl;
	CString str;
	CStringList *plstStr, *plstSearchStr;
	CPattern *pPattern, *pSearchPattern;
	bool bDefault;

	xmlapp=XmlDoc.FirstChildElement();

	for (xmlgroup=xmlapp->IterateChildren("group", 0); xmlgroup; xmlgroup=xmlapp->IterateChildren("group", xmlgroup))
	{
		plstStr=new CStringList();
		str=xmlgroup->ToElement()->Attribute("name");

		// check for dupe
		if (mapGroups.Lookup(str, (void *&)plstSearchStr))
		{
			msg("Duplicate group: %s\n", str);
			return false;
		}

		// add new group
		mapGroups.SetAt(str, plstStr);

		if (strcmp(xmlgroup->ToElement()->Attribute("default"), "on")==0)
			bDefault=true;
		else
			bDefault=false;

		for (xmlsearch=xmlgroup->IterateChildren("search", 0); xmlsearch; xmlsearch=xmlgroup->IterateChildren("search", xmlsearch))
		{
			if (xmlsearch)
			{
				str=ParseElement(xmlsearch);
				plstStr->AddTail(str);
				if (bDefault)
					lstDefaultGroups.AddTail(str);
			}
		}
	}

	for (xmlpattern=xmlapp->IterateChildren("pattern", 0); xmlpattern; xmlpattern=xmlapp->IterateChildren("pattern", xmlpattern))
	{
		// create new element
		pPattern = new CPattern();
		str=xmlpattern->ToElement()->Attribute("name");
		pPattern->SetName(str);

		// check for dupe
		if (mapPatterns.Lookup(str, (void *&)pSearchPattern))
		{
			msg("Duplicate pattern: %s\n", str);
			return false;
		}

		// add to list
		mapPatterns.SetAt(str, pPattern);
		
		str=xmlpattern->ToElement()->Attribute("default");
		if (str=="on")
		{
			pPattern->SetDefaultFlag(true);
			// add to default list
			lstDefaultPatterns.AddTail(pPattern);
		}

		xmlsearch=xmlpattern->FirstChild("search");

		if (xmlsearch)
		{
			str=ParseElement(xmlsearch);

			pPattern->SetSearchMask(str);

			xmlrepl=xmlpattern->FirstChild("replace");
			if (xmlrepl)
			{
				str=ParseElement(xmlrepl);
				pPattern->SetReplaceMask(str);
			}
		}
	}

	return true;
}

void CleanUpLists(void)
{
	POSITION pos;
	CString *s;
	ea_t addr;

	pos = mapPrefix.GetStartPosition();

	while (pos!=NULL)
	{
		mapPrefix.GetNextAssoc(pos, (void *&)addr, (void *&)s);
		delete s;
	}
	mapPrefix.RemoveAll();

	pos = mapPostfix.GetStartPosition();

	while (pos!=NULL)
	{
		mapPostfix.GetNextAssoc(pos, (void *&)addr, (void *&)s);
		delete s;
	}
	mapPostfix.RemoveAll();

}

void CleanUpPatterns(void)
{
	POSITION pos;
	CString str;
	CPattern *p;
	CStringList *lst;
	
	pos = mapPatterns.GetStartPosition();

	while (pos!=NULL)
	{
		mapPatterns.GetNextAssoc(pos, str, (void *&)p);
		delete p;
	}

	mapPatterns.RemoveAll();
	lstDefaultPatterns.RemoveAll();

	pos = mapGroups.GetStartPosition();
	while (pos != NULL)
	{
		mapGroups.GetNextAssoc(pos, str, (void *&)lst);
		delete lst;
	}
	mapGroups.RemoveAll();
	lstDefaultGroups.RemoveAll();
}

int idaapi Init(void)
{
//  if ( inf.filetype != f_PE ) return PLUGIN_SKIP;

  return PLUGIN_OK;
}

void idaapi Term(void)
{
}

void idaapi Run(int arg)
{
	msg("Cleaning code\n");
	msg("Parsing configuration file\n");
	
	char cFileName[MAX_PATH];
	qstrncpy(cFileName, idadir(NULL), MAX_PATH);
	qstrncat(cFileName, "\\plugins\\cleaner.xml", MAX_PATH);

	if (!XmlDoc.LoadFile(cFileName))
	{
		warning("Cannot open cleaner.xml\n");
		return;
	}

	if (!Parse())
	{
		msg("Parsing error!\n");
		return;
	}

	char init_dialog[] = 
		"STARTITEM 0\n"
		"Cleaner "
		CLEANER_VERSION
		"\n"
		"Type of clean:\n"		
		"<#Clean code at current address only#"
		"At current address :R>"

		"<#Clean code at range of addresses#"
		"At range :R>"

		"<#Rebuild code to text file#"
		"Rebuild code :R>>\n\n\n\n"

		"<#Reanalyse code (very slow!)#"
		"Reanalyse code:C>\n\n"

		"<#Do not clean non-code (data, undefined) areas#"
		"Skip non-code bytes:C>\n\n"

		"<#Show a lot of useless messages#"
		"Debug-mode:C>\n\n"

		"<#Write a lot of useless messages to CLEANER.LOG#"
		"Write debug messages to log:C>>\n\n"

		"<~S~tarting address:$:32:16::>\n"
		"<~E~nding address:$:32:16::>\n"
		"<~F~ile name:A:256:16::>\n";

	ea_t start_addr = get_screen_ea(); 
	ea_t end_addr = get_screen_ea();

	short sType;
	char buf[MAX_PATH];
	memset(buf, 0, MAX_PATH);

	int ok=AskUsingForm_c(init_dialog, &sType, &ulFlags, &start_addr, &end_addr, buf);
	if (ok)
	{
		show_wait_box("Cleaner: working...");

		switch(sType)
		{
		case 0:
			{
				if (Apply(start_addr))
					msg("succeeded\n");
				else
					msg("failed\n");
				break;
			}
		case 1:
			{
				for (ea_t i=start_addr; i<end_addr;i++)
				{
					if (wasBreak())
					{
						hide_wait_box();
						break;
					}

					Apply(i);
				}
				break;
			}
		case 2:
			{
				if (RebuildCode(start_addr, buf))
					msg("succeeded\n");
				else
					msg("failed\n");
				break;
			}
		}

		hide_wait_box();
	}

	CleanUpPatterns();
	CleanUpLists();
}

//--------------------------------------------------------------------------
char comment[] = "This is a cleaner plugin. It will clean code.";

char help[] =
        "A cleaner plugin module\n"
        "\n"
        "This module can clean you code according to .XML-definition.\n";


//--------------------------------------------------------------------------
// This is the preferred name of the plugin module in the menu system
// The preferred name may be overriden in plugins.cfg file

char wanted_name[] = "Cleaner";


// This is the preferred hotkey for the plugin module
// The preferred hotkey may be overriden in plugins.cfg file
// Note: IDA won't tell you if the hotkey is not correct
//       It will just disable the hotkey.

char wanted_hotkey[] = "Alt-z";


//--------------------------------------------------------------------------
//
//      PLUGIN DESCRIPTION BLOCK
//
//--------------------------------------------------------------------------
plugin_t PLUGIN =
{
  IDP_INTERFACE_VERSION,
  PLUGIN_UNL,           // plugin flags
  Init,                 // initialize

  Term,                 // terminate. this pointer may be NULL.

  Run,                  // invoke plugin

  comment,              // long comment about the plugin
                        // it could appear in the status line
                        // or as a hint

  help,                 // multiline help about the plugin

  wanted_name,          // the preferred short name of the plugin
  wanted_hotkey         // the preferred hotkey to run the plugin
};
