#include "pattern.h"

#include ".\TinyXML\TinyXML.h"
#include "2DArray.h"

#include <ida.hpp>
#include <idp.hpp>
#include <bytes.hpp>
#include <loader.hpp>
#include <kernwin.hpp>
#include <diskio.hpp>
#include <expr.hpp>
#include <xref.hpp>
#include <lines.hpp>
#include <name.hpp>
#include <ua.hpp>
#include <auto.hpp>

#define CLEANER_REANALYSE		0x00000001
#define	CLEANER_SKIP_NON_CODE	0x00000002
#define	CLEANER_DEBUG_MODE		0x00000004
#define	CLEANER_DEBUG_TOLOG		0x00000008

#define IsReanalyse(x)			((x & CLEANER_REANALYSE) == CLEANER_REANALYSE)
#define IsSkipNonCode(x)		((x & CLEANER_SKIP_NON_CODE) == CLEANER_SKIP_NON_CODE)
#define IsDebugMode(x)			((x & CLEANER_DEBUG_MODE) == CLEANER_DEBUG_MODE)
#define IsWriteLog(x)			((x & CLEANER_DEBUG_TOLOG) == CLEANER_DEBUG_TOLOG)

const char *JCC[16]={
		"jo     ",			// 70
		"jno    ",			// 71
		"jb     ",			// 72
		"jnb    ",			// 73
		"jz     ",			// 74
		"jnz    ",			// 75
		"jbe    ",			// 76
		"ja     ",			// 77
		"js     ",			// 78
		"jns    ",			// 79
		"jp     ",			// 7a
		"jnp    ",			// 7b
		"jl     ",			// 7c
		"jge    ",			// 7d
		"jle    ",			// 7e
		"jg     "			// 7f
};

class CStack : public CList<ea_t, const ea_t&>
{
public:
	// Add element to top of stack
	void Push( ea_t addr )
	{ AddHead( addr ); }

	// Peek at top element of stack
	ea_t Peek()
	{ return IsEmpty() ? BADADDR : GetHead(); }

	// Pop top element off stack
	ea_t Pop()
	{ return IsEmpty() ? BADADDR : RemoveHead(); }
};

BYTE ToByte(LPCSTR lpStr);
bool CheckSyntax(LPCSTR lpBuf);
void ReAnalyse(ea_t addr);
void ReAnalyseSmall(ea_t addr);
long ExecuteExpr(ea_t addr, LPCSTR lpExpr, CMapStringToPtr *pMap);
int CalcCounter(LPCSTR lpMask, int &iRes, CMapStringToPtr *pMap);
int CalcExprAndWrite(ea_t addr, LPCSTR lpMask, UINT &iLen, CMapStringToPtr *pMap);
bool CalcExprAndCompare(ea_t addr, LPCSTR lpMask, UINT &iMaskLen, UINT &iInstrLen, CMapStringToPtr *pMap);
long CalcExprAndCheck(ea_t addr, LPCSTR lpMask, UINT &iLen, CMapStringToPtr *pMap);

int ShowExpr(ea_t addr, LPCSTR lpMask, CMapStringToPtr *pMap);
int HideItem(ea_t &addr, LPCSTR lpMask, CMapStringToPtr *pMap, bool bStrong);
void HideByte(ea_t addr, bool bStrong);
int CalcHide(LPCSTR lpMask, int &iRes, unsigned char &bByte, CMapStringToPtr *pMap);
ea_t get_first_cref_to_high( ea_t ea );

ea_t SkipInstruction(ea_t addr);
int JumpExpr(ea_t &addr, LPCSTR lpMask, CMapStringToPtr *pMap);
ea_t SkipSingleSearch(ea_t addr, CString strSearch);
ea_t SkipSingleGroup(ea_t addr, CStringList *plst);
bool SkipGroup(ea_t &addr, bool bIsMultiple, LPCSTR lpBuf, UINT &iLen);

bool SkipPattern(ea_t &addr, bool bIsMultiple, LPCSTR lpBuf, UINT &iLen);
UINT SplitString(LPCSTR lpBuf, CStringList *lst);

int FillVars(ea_t addr, LPCSTR lpMask, UINT &iLen, CMapStringToPtr *pMap);
int SetVars(ea_t addr, LPCSTR lpMask, CMapStringToPtr *pMap);
ea_t Replace(ea_t addr, LPCSTR lpReplace, CMapStringToPtr *pMap);
ea_t IsMask(ea_t addr, LPCSTR lpMask, CMapStringToPtr *pMap);
ea_t Action(ea_t start_addr, CPattern *p);
bool Apply(ea_t start_addr);
ea_t SkipJump(ea_t addr);
BYTE IsConditionalJump(ea_t addr);
void DFS_WORK_TOPOLOGY(int iVal, LPBYTE lpVal, C2DArray *arr, CStack *stack);
UINT BuildGraph(ea_t cur_addr, ea_t from_addr, C2DArray *arr, CArray<ea_t, const ea_t&> *arrDone);
const char * GetJcc(BYTE bVal);
bool RebuildCode(ea_t addr, LPSTR lpFilename);
CString ParseElement(TiXmlNode *xmlnode);
bool Parse(void);
void CleanUpLists(void);
void CleanUpPatterns(void);

bool MakeComment(ea_t &addr, bool bIsRepeatable, LPCSTR lpBuf, UINT &iLen);
long ParseExpr(LPCSTR lpInBuf, LPSTR lpOutBuf, char cStopChar);
